'''
  $Id: bt_scan_cdx_json.py 158 2021-02-18 20:35:39Z cl_yubikey_2018a $
  
  Copyleft ITEG IT-Engineers GmbH 2022
'''

import getopt
#import glob
#import io
import json
#import os
import os.path
import shutil
#import subprocess
import sys

from tempfile import mkdtemp

from clazzes.bomtools import btlogging
from clazzes.bomtools import cyclone_dx
from clazzes.bomtools import dto
from clazzes.bomtools import helpers
from clazzes.bomtools import osv_report

log = btlogging.getLogger()


def describeUsage(argv0):
    usage = ""
    usage += "Usage: %s [options...] cdxJsonFile\n" % argv0
    usage += "\n"
    usage += "Scans a clondike_dx json file (.cdx.json file) for CVEs using osv-scanner by:\n"
    usage += "- calling osv-scanner -format json to get a vulnerability report\n"
    usage += "- parsing that vulnerability report and writing it to osv-report.csv, osv-report.html\n"
    usage += "\n"
    usage += "If license reporting is requested too /etc/bomtools/purl-license-mappings.conf\n"
    usage += "is evaluated for packages syft does not know the license of (which is all, for now).\n"
    #usage += "The 'purl' is constructed from 'pkg:deb/'+sourcePackageName+'@'+version.\n"
    usage += btlogging.getCommonUsageOptions()
    usage += "  Specific options are:\n"
    #usage += "    -e ecoSystem  default ecoSystem\n"
    #usage += "    -C csvFile   Write report in CSV format to specified file, '-' for stdout\n"
    #usage += "    -H htmlFile  Write report in HTML format to specified file, '-' for stdout, NYI\n"
    #usage += "    -k            Keep temporary directory, for investigation or follow-up tools\n"
    usage += "    -l            Report licenses too (licenses.csv, licenses.html)\n"
    usage += "    -L            Report licenses only\n"
    usage += "    -o outputDir  Reports directory, defaults to current directory\n"
    usage += "\n"
    usage += "cdxJsonFile must end in '.cdx.json' to satisfy osv-scanner.\n"
    usage += "\n"
    return usage


def run_bt_scan_cdx_json(argv0,args):
    
    try:
        opts, cdxJsonPaths = getopt.getopt(args, "dhlLo:qtv", ["help"])
    except getopt.GetoptError as err:
        print(err, file=sys.stderr) # will print something like "option -a not recognized"
        print(describeUsage(argv0), file=sys.stderr)
        sys.exit(64)

    btConfiguration = dto.BTConfiguration()

    reportOSV = True
    reportLicenses = False
    #keepTempDir = False
    #reportLicenses = False
    #licensesCsvFileName = "licenses.csv"
    #licensesHtmlFileName = "licenses.html"
  
    for o, a in opts:
        if o in ["-d", "-q", "-t", "-v"]:
            continue
        elif o in ["-h" , "--help"]:
            print(describeUsage(argv0))
            sys.exit(0)
        elif o in ["-l"]:
            reportLicenses = True
        elif o in ["-L"]:
            reportLicenses = True
            reportOSV = False
        elif o in ["-o"]:
            btConfiguration.reportsDirectory = a
        else:
            log.warning("Unhandled option: %s %s" %
                  (o, a))

    # chech jarPaths
    if len(cdxJsonPaths) != 1:
        log.error("Expecting exactly 1 .cdx.json file.")
        print(describeUsage(argv0), file=sys.stderr, flush=True)
        sys.exit(1)

    cdxJsonPath = cdxJsonPaths[0]

    if not cdxJsonPath.endswith(".cdx.json"):
        log.error("cdxJsonFile path does not end in '.cdx.json'.")
        print(describeUsage(argv0), file=sys.stderr, flush=True)
        sys.exit(1)

    # check osv-scanner
    osvScannerBin = None
    if reportOSV: 
        osvScannerBin = helpers.checkOsvScannerBin()

    # check missing path parts
    if not os.path.isfile(cdxJsonPath):
        log.error("Specified cdxJsonFile '%s' is not present." % cdxJsonPath)
        sys.exit(1)

    os.makedirs(btConfiguration.reportsDirectory, exist_ok=True)
    
    log.info("Analysing %s ..." % cdxJsonPath)

    # temp dir
    #tempDirPath = mkdtemp(suffix=None, prefix=None, dir=None)
    #log.debug("Created temporary directory: '%s'" % tempDirPath)
        
    finalRc = 0
    btConfiguration.defaultEcoSystem = "?"
    #nrOsvVulnerablePackages = 0
    #nrOsvVulnerabilities = 0
    try:
        
        if reportLicenses:
            #  calling syft -o cyclonedx-json to create a BOM (bill of materials)
            btConfiguration.syftBomJsonPath = cdxJsonPath
            #btConfiguration.htmlHints = "<p></p>"
            cyclone_dx.loadSyftBomAndReportLicenses(btConfiguration=btConfiguration)
        else:
            log.notice("License reporting not requested, not creating or loading BOM")

        if reportOSV:
            #  calling osv-scanner -format json to get a vulnerability report
            osvScannerCmdParts = [osvScannerBin, "-format", "json", cdxJsonPath]
            #btConfiguration.htmlHints = "<p></p>"
            osv_report.performOsvScannerAndReport(btConfiguration=btConfiguration, osvScannerCmdParts=osvScannerCmdParts)
        else:
            log.notice("OSV reporting suppressed, not calling osv-scanner")

    except Exception as exception:
        log.error("Caught an error: %s" % exception)
        finalRc = 2
        raise(exception)

    #finally:
    #    #  deleting the temporary directory
    #    if keepTempDir:
    #        log.notice("Keeping temporary directory '%s'" % tempDirPath)
    #    else:
    #        log.debug("Removing temporary directory '%s' ..." % tempDirPath)
    #        shutil.rmtree(tempDirPath)
            
    log.debug("Final exit code: %d" % finalRc)
    sys.exit(finalRc)
