'''
  $Id: bt_scan_jars.py 158 2021-02-18 20:35:39Z cl_yubikey_2018a $
  
  Copyleft ITEG IT-Engineers GmbH 2022
'''

import getopt
import glob
#import io
import json
#import os
import os.path
import shutil
#import subprocess
import sys
from cvss import cvss3

from tempfile import mkdtemp

from clazzes.bomtools import cyclone_dx
from clazzes.bomtools import dto
from clazzes.bomtools import helpers
from clazzes.bomtools import btlogging
from clazzes.bomtools import osv_report

log = btlogging.getLogger()


def describeUsage(argv0):
    usage = ""
    usage += "Usage: %s [options...] jarFileOrDirectory...\n" % argv0
    usage += "\n"
    usage += "Scans JAR files for CVEs using syft and osv-scanner by:\n"
    usage += "- filling a temporary directory with links to all specified JAR files,\n"
    usage += "  to restrict scans to JAR files only (on purpose)\n"
    usage += "- calling syft -o cyclonedx-json to create a BOM (bill of materials)\n"
    usage += "  in jars.cdx.json in that directory\n"
    usage += "- calling osv-scanner -format json to get a vulnerability report\n"
    usage += "- parsing that vulnerability report and writing it to java-osv-report.csv, java-osv-report.html\n"
    usage += "- deleting the temporary directory\n"
    usage += "\n"
    usage += "If license reporting is requested too /etc/bomtools/purl-license-mappings.conf\n"
    usage += "is evaluated for packages syft does not know the license of.\n"
    usage += "The 'purl' is constructed from 'pkg:maven/'+groupId+'/'+artifactId+'@'+version.\n"
    usage += btlogging.getCommonUsageOptions()
    usage += "  Specific options are:\n"
    #usage += "    -C csvFile   Write report in CSV format to specified file, '-' for stdout\n"
    #usage += "    -H htmlFile  Write report in HTML format to specified file, '-' for stdout, NYI\n"
    usage += "    -i            Ignore (not abort over) missing files or directories\n"
    usage += "    -k            Keep temporary directory, for investigation or follow-up tools\n"
    usage += "    -l            Report licenses too (java-licenses.csv, java-licenses.html)\n"
    usage += "    -L            Report licenses only\n"
    usage += "    -o outputDir  Reports directory, defaults to current directory\n"
    usage += "\n"
    return usage


def run_bt_scan_jars(argv0,args):
    
    try:
        opts, jarPaths = getopt.getopt(args, "dhiklLo:qtv", ["help"])
    except getopt.GetoptError as err:
        print(err, file=sys.stderr) # will print something like "option -a not recognized"
        print(describeUsage(argv0), file=sys.stderr)
        sys.exit(64)

    btConfiguration = dto.BTConfiguration()

    ignoreMissingFiles = False
    reportOSV = True
    reportLicenses = False
    keepTempDir = False

    btConfiguration.licensesCsvFileName = "java-licenses.csv"
    btConfiguration.licensesHtmlFileName = "java-licenses.html"
    btConfiguration.osvCsvFileName = "java-osv-report.csv"
    btConfiguration.osvHtmlFileName = "java-osv-report.html"

    for o, a in opts:
        if o in ["-d", "-q", "-t", "-v"]:
            continue
        elif o in ["-h" , "--help"]:
            print(describeUsage(argv0))
            sys.exit(0)
        elif o in ["-i"]:
            ignoreMissingFiles = True
        elif o in ["-k"]:
            keepTempDir = True
        elif o in ["-l"]:
            reportLicenses = True
        elif o in ["-L"]:
            reportLicenses = True
            reportOSV = False
        elif o in ["-o"]:
            btConfiguration.reportsDirectory = a
        else:
            log.warning("Unhandled option: %s %s" %
                  (o, a))

    # check syft
    # we need the syft output, to feed osv-scanner
    syftBin = helpers.checkSyftBin()

    # check osv-scanner
    osvScannerBin = None
    if reportOSV: 
        osvScannerBin = helpers.checkOsvScannerBin()

    # chech jarPaths
    if len(jarPaths) < 1:
        log.error("Expecting at least one jar file or directory.")
        print(describeUsage(argv0), file=sys.stderr, flush=True)
        sys.exit(1)

    os.makedirs(btConfiguration.reportsDirectory, exist_ok=True)
    licensesCsvFileName = "%s/licenses.csv" % btConfiguration.reportsDirectory
    licensesHtmlFileName = "%s/licenses.html" % btConfiguration.reportsDirectory

    jarFilePaths = []
    for jarPath in jarPaths:
        # Try to nominalize relative paths to absolute for easier handling
        absJarPath = os.path.abspath(jarPath)
        
        if os.path.isfile(absJarPath):
            log.debug("Found JAR file %s" % absJarPath)
            jarFilePaths.append(absJarPath)
        elif os.path.isdir(absJarPath):
            log.debug("Found JAR directory %s" % absJarPath)
            for jarFileName in glob.iglob(f"{absJarPath}/**/*.jar", recursive=True):
                absJarFilePath = os.path.abspath(jarFileName)
                log.debug("Found JAR file %s" % absJarFilePath)
                jarFilePaths.append(absJarFilePath)
        elif ignoreMissingFiles:
            log.debug("Ignoring missing jarPath '%s'" % jarPath)
        else:
            log.error("Aborting due to missing jarPath '%s'" % jarPath)
            sys.exit(1)

    log.info("Analysing %d jar files ..." % len(btConfiguration.jarFiles))
    
    # temp dir
    #tempDirPath = tempfile.TemporaryDirectory()
    tempDirPath = mkdtemp(suffix=None, prefix=None, dir=None)
    log.debug("Created temporary directory: '%s'" % tempDirPath)

    jarsLinksPath = "%s/jars" % tempDirPath
    log.debug("Creating temporary subdirectory for '*.jar' links: '%s'" % jarsLinksPath)
    os.makedirs(jarsLinksPath)

    finalRc = 0
    btConfiguration.defaultEcoSystem = "Maven"

    try:

        #  filling a temporary directory with links to all specified JAR files
        for jarFilePath in sorted(jarFilePaths):
            targetFilename = "%s/%s" % (jarsLinksPath, jarFilePath[1:].replace('/', '_'))
            if os.path.exists(targetFilename):
                log.debug("Not replacing existing %s with symlink to %s" % (targetFilename, jarFilePath))
            else:
                log.debug("Symlinking %s to %s" % (targetFilename, jarFilePath))
                os.symlink(jarFilePath, targetFilename)
    
        #  calling syft -o cyclonedx-json to create a BOM (bill of materials)
        btConfiguration.syftBomJsonPath = "%s/syft-bom.cdx.json"%tempDirPath
        syftCmdParts = [syftBin, "scan", jarsLinksPath, "-o", "cyclonedx-json=%s" % btConfiguration.syftBomJsonPath, "--base-path", "/" ]
        syftRc = helpers.performDebugLogged(syftCmdParts)
        if syftRc != 0:
            log.error("Got bad RC %d from %s" % (syftRc, " ".join(syftCmdParts)))
            raise Exception("Failed to perform %s" % (syftRc, " ".join(syftCmdParts)))
    
        if reportLicenses:
            log.notice("License reporting requested, loading BOM ...")
            btConfiguration.htmlHints = "" # <p>Package names refer to binary packages.</p>"
            cyclone_dx.loadSyftBomAndReportLicenses(btConfiguration=btConfiguration)
        else:
            log.notice("License reporting not requested, not loading BOM")
    
        if reportOSV:
            #  calling osv-scanner -format json to get a vulnerability report
            osvScannerCmdParts = [osvScannerBin, "-format", "json", btConfiguration.syftBomJsonPath]
            btConfiguration.htmlHints = "" # <p>Package names refer to source packages.</p>"
            osv_report.performOsvScannerAndReport(btConfiguration=btConfiguration, osvScannerCmdParts=osvScannerCmdParts)
        else:
            log.notice("OSV reporting suppressed, not calling osv-scanner")
            

    except Exception as exception:
        log.error("Caught an error: %s" % exception)
        finalRc = 2
        raise(exception)

    finally:
        #  deleting the temporary directory
        if keepTempDir:
            log.notice("Keeping temporary directory '%s'" % tempDirPath)
        else:
            log.debug("Removing temporary directory '%s' ..." % tempDirPath)
            shutil.rmtree(tempDirPath)
    
    log.debug("Final exit code: %d" % finalRc)
    sys.exit(finalRc)
    
    

