'''
  $Id: bt_scan_npm_jsons.py 158 2021-02-18 20:35:39Z cl_yubikey_2018a $
  
  Copyleft ITEG IT-Engineers GmbH 2022
'''

import getopt
#import glob
#import io
import json
#import os
import os.path
import shutil
#import subprocess
import sys

from tempfile import mkdtemp

from clazzes.bomtools import btlogging
from clazzes.bomtools import cyclone_dx
from clazzes.bomtools import dto
from clazzes.bomtools import helpers
from clazzes.bomtools import osv_report

log = btlogging.getLogger()


def describeUsage(argv0):
    usage = ""
    usage += "Usage: %s [options...] npmCdxJsonFile [npmLicensesFile]\n" % argv0
    usage += "\n"
    usage += "Scans a clondike_dx json file (.cdx.json file) for CVEs using osv-scanner by:\n"
    usage += "- calling osv-scanner -format json to get a vulnerability report\n"
    usage += "- parsing that vulnerability report and writing it to npm-osv-report.csv, npm-osv-report.html\n"
    usage += "\n"
    usage += "If license reporting is requested too /etc/bomtools/purl-license-mappings.conf\n"
    usage += "is evaluated for packages syft does not know the license of (which is all, for now).\n"
    #usage += "The 'purl' is constructed from 'pkg:deb/'+sourcePackageName+'@'+version.\n"
    usage += btlogging.getCommonUsageOptions()
    usage += "  Specific options are:\n"
    #usage += "    -e ecoSystem  default ecoSystem\n"
    #usage += "    -C csvFile   Write report in CSV format to specified file, '-' for stdout\n"
    #usage += "    -H htmlFile  Write report in HTML format to specified file, '-' for stdout, NYI\n"
    #usage += "    -k            Keep temporary directory, for investigation or follow-up tools\n"
    usage += "    -l            Report licenses too (npm-licenses.csv, npm-licenses.html),\n"
    usage += "                  requires a npmLicensesFile as json by yarn.\n"
    usage += "    -L            Report licenses only\n"
    usage += "    -o outputDir  Reports directory, defaults to current directory\n"
    usage += "\n"
    usage += "npmCdxJsonFile must end in '.cdx.json' to satisfy osv-scanner.\n"
    usage += "\n"
    #usage += "If npmLicensesFile is specified -l is automatically triggered.\n"
    #usage += "\n"
    return usage


def run_bt_scan_npm_jsons(argv0,args):
    
    try:
        opts, npmJsonPaths = getopt.getopt(args, "dhlLo:qtv", ["help"])
    except getopt.GetoptError as err:
        print(err, file=sys.stderr) # will print something like "option -a not recognized"
        print(describeUsage(argv0), file=sys.stderr)
        sys.exit(64)

    btConfiguration = dto.BTConfiguration()

    reportOSV = True
    reportLicenses = False
    #keepTempDir = False
    #reportLicenses = False
    btConfiguration.licensesCsvFileName = "npm-licenses.csv"
    btConfiguration.licensesHtmlFileName = "npm-licenses.html"
    btConfiguration.osvCsvFileName = "npm-osv-report.csv"
    btConfiguration.osvHtmlFileName = "npm-osv-report.html"
  
    for o, a in opts:
        if o in ["-d", "-q", "-t", "-v"]:
            continue
        elif o in ["-h" , "--help"]:
            print(describeUsage(argv0))
            sys.exit(0)
        elif o in ["-l"]:
            reportLicenses = True
        elif o in ["-L"]:
            reportLicenses = True
            reportOSV = False
        elif o in ["-o"]:
            btConfiguration.reportsDirectory = a
        else:
            log.warning("Unhandled option: %s %s" %
                  (o, a))

    # chech jarPaths
    if len(npmJsonPaths) < 1:
        log.error("No npmCdxJsonFile file (.cdx.json) or npmLicensesFile file specified.")
        print(describeUsage(argv0), file=sys.stderr, flush=True)
        sys.exit(1)

    if len(npmJsonPaths) > 2:
        log.error("Expecting not more than one npmCdxJsonFile file (.cdx.json) and one npmLicensesFile file.")
        print(describeUsage(argv0), file=sys.stderr, flush=True)
        sys.exit(1)

    cdxJsonPath = npmJsonPaths[0]

    if not cdxJsonPath.endswith(".cdx.json"):
        log.error("cdxJsonFile path does not end in '.cdx.json'.")
        print(describeUsage(argv0), file=sys.stderr, flush=True)
        sys.exit(1)

    npmLicensesJsonPath = None
    
    if reportLicenses and len(npmJsonPaths) == 1:
        log.error("Reporting licenses requires a npmLicensesFile specified as 2nd positional parameter.")
        print(describeUsage(argv0), file=sys.stderr, flush=True)
        sys.exit(1)
        
    if len(npmJsonPaths) == 2:
        npmLicensesJsonPath = npmJsonPaths[1]
        #reportLicenses = True

    # check osv-scanner
    osvScannerBin = None
    if reportOSV: 
        osvScannerBin = helpers.checkOsvScannerBin()

    # check missing path parts
    if not os.path.isfile(cdxJsonPath):
        log.error("Specified cdxJsonFile '%s' is not present." % cdxJsonPath)
        sys.exit(1)

    if npmLicensesJsonPath != None and not os.path.isfile(npmLicensesJsonPath):
        log.error("Specified npmCdxJsonFile '%s' is not present." % npmLicensesJsonPath)
        sys.exit(1)

    os.makedirs(btConfiguration.reportsDirectory, exist_ok=True)
    
    log.info("Analysing %s ..." % cdxJsonPath)

    # temp dir
    #tempDirPath = mkdtemp(suffix=None, prefix=None, dir=None)
    #log.debug("Created temporary directory: '%s'" % tempDirPath)
        
    finalRc = 0
    btConfiguration.defaultEcoSystem = "npm"
    #nrOsvVulnerablePackages = 0
    #nrOsvVulnerabilities = 0
    try:
        
        if reportLicenses:
            # create PURL mappings
            with open(npmLicensesJsonPath, 'r') as jsonFile:
                npmLicenses = json.load(jsonFile)
                for npmPackageInfo in npmLicenses["data"]["body"]:
                    packageName       = npmPackageInfo[0]
                    packageVersion    = npmPackageInfo[1]
                    packageLicense    = npmPackageInfo[2]
                    #packageUrl        = npmPackageInfo[3]
                    #packageVendorUrl  = npmPackageInfo[4]
                    #packageVendorName = npmPackageInfo[5]
                    packagePurl = "pkg:npm/%s@%s" % (packageName, packageVersion)
                    log.debug("Creating PURLLicenseMapping from '%s' to '%s'" % (packagePurl, packageLicense))
                    purlLicenseMapping = dto.PURLLicenseMapping(purlPattern=packagePurl, licenseInfo=packageLicense)
                    btConfiguration.purlLicenseMappings.append(purlLicenseMapping)
                    # verions mismatch for now, so add another one ...
                    #packagePurlNoVersion = "pkg:npm/%s@" % packageName
                    #log.debug("Creating 2nd PURLLicenseMapping from '%s' to '%s'" % (packagePurlNoVersion, packageLicense))
                    #purlLicenseMappingNoVersion = dto.PURLLicenseMapping(purlPattern=packagePurlNoVersion, licenseInfo=packageLicense)
                    #btConfiguration.purlLicenseMappings.append(purlLicenseMappingNoVersion)
                    # some artifacts are missing, so add another one ...
                    #packagePurlNoArtifact = "pkg:npm/%s" % packageName.split("/")[0]
                    #log.debug("Creating 3rd PURLLicenseMapping from '%s' to '%s'" % (packagePurlNoArtifact, packageLicense))
                    #purlLicenseMappingNoArtifact = dto.PURLLicenseMapping(purlPattern=packagePurlNoArtifact, licenseInfo=packageLicense)
                    #btConfiguration.purlLicenseMappings.append(purlLicenseMappingNoArtifact)
                
            btConfiguration.syftBomJsonPath = cdxJsonPath
            #btConfiguration.htmlHints = "<p></p>"
            #  calling syft -o cyclonedx-json to create a BOM (bill of materials)
            cyclone_dx.loadSyftBomAndReportLicenses(btConfiguration=btConfiguration)
        else:
            log.notice("License reporting not requested, not creating or loading BOM")

        if reportOSV:
            #  calling osv-scanner -format json to get a vulnerability report
            osvScannerCmdParts = [osvScannerBin, "-format", "json", cdxJsonPath]
            #btConfiguration.htmlHints = "<p></p>"
            osv_report.performOsvScannerAndReport(btConfiguration=btConfiguration, osvScannerCmdParts=osvScannerCmdParts)
        else:
            log.notice("OSV reporting suppressed, not calling osv-scanner")

    except Exception as exception:
        log.error("Caught an error: %s" % exception)
        finalRc = 2
        raise(exception)

    #finally:
    #    #  deleting the temporary directory
    #    if keepTempDir:
    #        log.notice("Keeping temporary directory '%s'" % tempDirPath)
    #    else:
    #        log.debug("Removing temporary directory '%s' ..." % tempDirPath)
    #        shutil.rmtree(tempDirPath)
            
    log.debug("Final exit code: %d" % finalRc)
    sys.exit(finalRc)
