'''
  Copyleft ITEG IT-Engineers GmbH 2023
'''

import json
#import pathlib

from datetime import datetime
from typing import List, Optional, TypedDict

from clazzes.bomtools import btlogging
from clazzes.bomtools import dto

log = btlogging.getLogger()


# data classes für CycloneDX' JSON structure


class Hash(TypedDict):
    alg: str
    content: str


class ExternalReference(TypedDict):
    url: str
    hashes: List[Hash]
    type: str


class LicenseLicense(TypedDict):
    name: Optional[str]
    id: Optional[str]


class LicenseElement(TypedDict):
    license: LicenseLicense


class Property(TypedDict):
    name: str
    value: str


class ComponentElement(TypedDict):
    bom_ref: str
    type: str
    name: str
    version: str
    cpe: str
    purl: str
    properties: List[Property]
    group: Optional[str]
    licenses: Optional[List[LicenseElement]]
    external_references: Optional[List[ExternalReference]]


class MetadataComponent(TypedDict):
    bom_ref: str
    type: str
    name: str


class Tool(TypedDict):
    vendor: str
    name: str
    version: str


class Metadata(TypedDict):
    timestamp: datetime
    tools: List[Tool]
    component: MetadataComponent


class CycloneDX(TypedDict):
    schema: str
    bom_format: str
    spec_version: str
    serial_number: str
    version: int
    metadata: Metadata
    components: List[ComponentElement]


def loadCycloneDxBomJson(jsonFilePath:str = None):
    with open(jsonFilePath, 'r') as jsonFile:
        bom: CycloneDX = json.load(jsonFile)
        return bom

def loadSyftBomAndReportLicenses(btConfiguration:dto.BTConfiguration=None):
    log.notice("License reporting requested, loading BOM ...")
    # based on https://gitlab.intra.iteg.at/dds/dds-service/-/merge_requests/576/diffs
    bom = loadCycloneDxBomJson(btConfiguration.syftBomJsonPath)
    #components = []
    #bom["components"].sort(key=lambda component: component["purl"])
    #bom["components"].sort(key=lambda component: "%s_%s" % (component["name"] != None, component["version"]))
    # does not compute: components = filter(lambda component: component["name"] != None and component["version"] != None, bom["components"])
    # does not compute: components = [component for component in bom["components"] if (component.has_key('name')) and (component.has_key('version'))]
    components = [component for component in bom["components"] if component["type"] != "file"]
    components.sort(key=lambda component: "%s_%s" % (component["name"] != None, component["version"]))
    # loading worked, start files
    csvFileHandle = open("%s/%s" % (btConfiguration.reportsDirectory, btConfiguration.licensesCsvFileName), 'w')
    csvFileHandle.write('Name,Version,License,Virtual Path\n')
    # HTML file header
    htmlFileHandle = open("%s/%s" % (btConfiguration.reportsDirectory, btConfiguration.licensesHtmlFileName), 'w')
    htmlFileHandle.write("<h1>Licenses</h1>\n")
    if len(btConfiguration.htmlHints) > 1:
        htmlFileHandle.write(btConfiguration.htmlHints)
    htmlFileHandle.write('<table border="1" rules="all">\n  <thead><tr><th>Name</th><th>Version</th><th>License</th><th>Virtual Path</th></tr></thead>\n<tbody>\n')
    
    #for component in bom["components"]:
    for component in components:
        packageFoundBy = next(
            (
                m["value"]
                for m in component["properties"]
                if m["name"] == "syft:package:foundBy"
            ),
            "",
        )
        name = component["name"]
        version = component["version"]
        purl = None
        if  (packageFoundBy == "java-cataloger") or  (packageFoundBy == "java-archive-cataloger"):
            purl = component["purl"]
        elif (packageFoundBy == "dpkgdb-cataloger") or (packageFoundBy == "dpkg-db-cataloger"):
            sourcePackageName = next(
                (
                    m["value"]
                    for m in component["properties"]
                    if m["name"] == "syft:metadata:source"
                ),
                name,
            )
            purl = 'pkg:deb/%s@%s' % (sourcePackageName, version)
        elif packageFoundBy == "javascript-lock-cataloger":
            purl = component["purl"].replace("%40","@")
        elif (component["purl"] != None) and component["purl"].startswith("pkg:npm/"):
            # source is 'npm run sbom'
            purl = component["purl"]
        elif packageFoundBy == None:
            log.error("loadSyftBomAndReportLicenses(): Ignoring package %s v. %s without syft:package:foundBy property" % (name, version))
            continue
        else:
            log.error("loadSyftBomAndReportLicenses(): Ignoring package %s v. %s with unknown syft:package:foundBy property '%s'" % (name, version, packageFoundBy))
            continue
        #log.debug("component: %s" % component)
        #log.debug("type of component.licenses: %s" % type(component.get("licenses")))
        licenses = []
        for i in component.get("licenses", []):
            if i.get("license") != None:
                licenses.append(i["license"].get("id", i["license"].get("name", "")))
            elif i.get("expression"):
                licenses.append(i.get("expression"))
        #log.debug("past license loop")
        if licenses:
            license = ", ".join(licenses)
        else:
            #licenses = [v.licenseInfo for (k, v) in btConfiguration.purlLicenseMappings.items() if purl.startswith(k)]
            license = next(
                (
                    mapping.licenseInfo
                    for mapping in btConfiguration.purlLicenseMappings
                    if purl.startswith(mapping.purlPattern)
                ),
                ""
            )
        virtualPath = next(
            (
                m["value"]
                for m in component["properties"]
                if (m["name"] == "syft:metadata:virtualPath") or (m["name"] == "cdx:npm:package:path")
            ),
            "",
        )
        if len(license) == 0:
            license = "- (UNKNOWN for PURL %s)" % purl
        #print(f"|{purl}|{license}|{virtualPath}|")
        csvFileHandle.write('%s,%s,"%s",%s\n' % (
            name, version, license.replace('"', '\\"'), virtualPath))
        htmlFileHandle.write('  <tr><td>%s</td><td>%s</td><td>%s</td><td>%s</td></tr>\n' % (
            name, version, license, virtualPath))
    
    csvFileHandle.close()
    # HTML file footer?
    htmlFileHandle.write("</tbody></table>\n")
    if len(btConfiguration.htmlHints) > 1:
        htmlFileHandle.write(btConfiguration.htmlHints)
    htmlFileHandle.write("<p>Created by bomtools at %s</p>" % datetime.now().astimezone().isoformat(sep=" ", timespec="seconds"))
    htmlFileHandle.close()
