'''
  $Id: helpers.py 192 2021-03-11 15:19:05Z cl_yubikey_2018a $
  
  Copyleft ITEG IT-Engineers GmbH 2022
'''

import fileinput
#import ipaddress
#import json
#import os
import re
import shutil
#import socket
import subprocess
import sys

from cvss import CVSS2, CVSS3, CVSS4
from cvss.exceptions import CVSS2Error, CVSS3Error, CVSS4Error

from clazzes.bomtools import btlogging

log = btlogging.getLogger()


def formatFileSize(size:int = -1):
    if size == None:
        return "- (None)"
    if size < 0:
        return "- (< 0)"
    if size < 1024:
        return "%d B" % size
    if size < 1048576:
        return "%d KB" % (size/1024)
    if size < 1073741824:
        return "%d MB" % (size/1048576)
    if size < 1099511627776:
        return "%d GB" % (size/1073741824)
    return "%1.2f TB" % (size/1099511627776)


def parseFileSize(sizeString:str = ""):
    if sizeString == None:
        return -1
    if len(sizeString) == 0:
        return -2
    reInt = re.compile(r"^\d+$")
    #matchInt = reInt.search(sizeString)
    if reInt.fullmatch(sizeString):
        return int(sizeString)
        
    reIntUnit = re.compile(r"^(\d+)([mgtMGT])$")
    matchIntUnit = reIntUnit.fullmatch(sizeString)
    if matchIntUnit == None:
        return -3
    sizeInt = int(matchIntUnit.group(1))
    sizeBytes = -4
    if matchIntUnit.group(2) in "mM":
        sizeBytes = sizeInt * 1048576
    elif matchIntUnit.group(2) in "gG":
        sizeBytes = sizeInt * 1073741824
    elif matchIntUnit.group(2) in "tT":
        sizeBytes = sizeInt * 1099511627776
    #log.debug("parseFileSize(): Computed '%s' to %d" % (
    #    sizeString, sizeBytes))
    return sizeBytes


def readConf(confFileName:str = ""):
    if len(confFileName) == 0:
        raise Exception("ERROR: readConf() called with empty confFileName")
    conf = {}
    lineNr = 0
    for line in fileinput.input(files=(confFileName)):
        lineNr += 1
        if line[-1] == '\n':
            line = line[:-1]
        #log.debug("readConf(%s:%d): line='%s'" % (confFileName, lineNr, line))
        if len(line) == 0:
            #log.debug("readConf(%s:%d): Ignoring emtpy line" % (confFileName, lineNr))
            continue
        if line[0] in "#;":
            #log.debug("readConf(%s:%d): Comment line" % (confFileName, lineNr))
            continue
        lineParts = line.split('=')
        if len(lineParts) < 1:
            log.warning("readConf(%s:%d): Ignoring invalid line without '=': '%s'" % (
                confFileName, lineNr, line))
            continue
        if len(lineParts[0]) == 0:
            log.warning("readConf(%s:%d): Ignoring invalid line without option name: '%s'" % (
                confFileName, lineNr, line))
            continue
        name = lineParts[0].strip()
        if len(name) == 0:
            log.warning("readConf(%s:%d): Ignoring invalid line with empty option name: '%s'" % (
                confFileName, lineNr, line))
            continue
        #log.debug("readConf(%s:%d): Trimmed name: '%s'" % (confFileName, lineNr, name))
        value = "=".join(lineParts[1:]).strip()
        # empty value is allowed
        #log.debug("readConf(%s:%d): Trimmed value: '%s'" % (confFileName, lineNr, value))
        if len(value) > 1 and value[0] == '"' and value[-1] == '"':
            value = value[1:-1]
        for key in conf.keys():
            value = value.replace("$%s" % key, conf[key]).replace("${%s}" % key, conf[key])
        #log.debug("readConf(%s:%d): $-replaced value: '%s'" % (confFileName, lineNr, value))
        conf[name] = value
    
    return conf


def readMappingList(confFileName:str = ""):
    if len(confFileName) == 0:
        raise Exception("ERROR: readMappingList() called with empty confFileName")
    mappings = []
    lineNr = 0
    for line in fileinput.input(files=(confFileName)):
        lineNr += 1
        if line[-1] == '\n':
            line = line[:-1]
        #log.debug("readConf(%s:%d): line='%s'" % (confFileName, lineNr, line))
        if len(line) == 0:
            #log.debug("readConf(%s:%d): Ignoring emtpy line" % (confFileName, lineNr))
            continue
        if line[0] in "#;":
            #log.debug("readConf(%s:%d): Comment line" % (confFileName, lineNr))
            continue
        lineParts = line.split('=')
        if len(lineParts) < 1:
            log.warning("readMappingList(%s:%d): Ignoring invalid line without '=': '%s'" % (
                confFileName, lineNr, line))
            continue
        if len(lineParts[0]) == 0:
            log.warning("readMappingList(%s:%d): Ignoring invalid line without pattern: '%s'" % (
                confFileName, lineNr, line))
            continue
        pattern = lineParts[0].strip()
        if len(pattern) == 0:
            log.warning("readMappingList(%s:%d): Ignoring invalid line with empty pattern: '%s'" % (
                confFileName, lineNr, line))
            continue
        #log.debug("readConf(%s:%d): Trimmed name: '%s'" % (confFileName, lineNr, name))
        mapping = "=".join(lineParts[1:]).strip()
        # empty value is allowed
        #log.debug("readConf(%s:%d): Trimmed value: '%s'" % (confFileName, lineNr, value))
        if len(mapping) > 1 and mapping[0] == '"' and mapping[-1] == '"':
            mapping = mapping[1:-1]
        #log.debug("readConf(%s:%d): $-replaced value: '%s'" % (confFileName, lineNr, value))
        mappings.append([pattern, mapping])
    
    return mappings


def backcheckOnProblem(rc:int=-1, cmdParts:[]=[], problem:str=""):
    if len(cmdParts) > 0:
        print("ATTENTION: The former command,\n  %s\nreturned the bad exit code %d." % (
            " ".join(cmdParts), rc), file=sys.stderr, flush=True)
    else:
        print("ATTENTION: The former command returned the bad exit code %d." %
              rc, file=sys.stderr, flush=True)

    if len(problem) > 0:
        print("PROBLEM description: %s" %
              problem, file=sys.stderr, flush=True)

    choice = input("Please try to correct the problem and re-execute the command manually in another terminal.\n"
                   "If your are You sure it is ok to continue, enter 'yes' ... ")
    if choice != "yes":
        log.error("You chose to ABORT halfway because of the problem mentioned above.")
        sys.exit(1)

    log.warning("Continuing on user's request")    
    
    
def performDebugLogged(cmdParts:list=[], collectOutput=False):
    if len(cmdParts) == 0:
        raise Exception("ERROR: performDebugLogged() called with empty cmdParts")
    log.debug("Running\n  %s" % " ".join(cmdParts))
    cp = None
    if collectOutput:
        cp = subprocess.run(cmdParts, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
    else:
        cp = subprocess.run(cmdParts)
    rc = cp.returncode
    if rc != 0:
        log.debug("Got bad exitcode %d from %s" % (
            rc,  " ".join(cmdParts)))
    else:
        log.debug("Returncode was: %d" % rc)
    if collectOutput:
        return rc, cp.stdout, cp.stderr
    return rc


def checkOsvScannerBin():
    osvScannerBin = shutil.which("osv-scanner")
    if osvScannerBin is None:
        log.error("Required tool 'osv-scanner' is not installed.")
        sys.exit(1)
    else:
        log.debug("Found osv-scanner at %s" % osvScannerBin)
        return osvScannerBin

def checkSyftBin():
    syftBin = shutil.which("syft")
    if syftBin is None:
        log.error("Required tool 'syft' is not installed.")
        sys.exit(1)
    else:
        log.debug("Found syft at %s" % syftBin)
        return syftBin

def getCvssScore(vector_or_json: str | dict[str, str]):
    """
        Determines CVSS type and vector based on input
        Input is either the vector string directly (in which case CVSS_V3 is assumed) or the value found in
        the "severity" field of an obj in the "vulnerabilities" field of the osv json
    """
    log.debug("Attempting to parse score ${vector_or_json}")
    if isinstance(vector_or_json, str):
        cvss_type = "CVSS_V3"
        cvss_vector = vector_or_json
    elif isinstance(vector_or_json, dict) and "type" in vector_or_json and "score" in vector_or_json:
        cvss_type = vector_or_json["type"]
        cvss_vector = vector_or_json["score"]
    else:
        log.warning("CVSS score has invalid input format")
        return None

    cvss_map = {
        "CVSS_V2": (CVSS2, CVSS2Error),
        "CVSS_V3": (CVSS3, CVSS3Error),
        "CVSS_V4": (CVSS4, CVSS4Error),
    }

    if cvss_type in cvss_map:
        try:
            cvss_func, error_class = cvss_map[cvss_type]
            cvss = cvss_func(cvss_vector)
        except error_class as e:
            log.warning(f"Could not parse {cvss_vector}: {e}")
            return None
    else:
        log.warning(f"Unsupported CVSS type: {cvss_type}")
        return None

    numerical_score = cvss.scores()[0] if cvss.scores() else None
    log.debug(f"CVSS score of {vector_or_json} is {numerical_score if numerical_score is not None else 'not specified'}")
    return numerical_score