'''
  Copyleft ITEG IT-Engineers GmbH 2023
'''

import json
#import pathlib

from datetime import datetime, timezone
#from typing import List, Optional, TypedDict

from clazzes.bomtools import btlogging
from clazzes.bomtools import dto
from clazzes.bomtools import helpers

log = btlogging.getLogger()


# data classes für CycloneDX' JSON structure

'''
class Vulnerability(TypedDict):
    schema_version: Optional[str]
    id: str
    summary: str
    
class Package(TypedDict):
    package: str
    name: str
    version: str
    ecosystem: str
    vulnerabilities: List[Vulnerability]

class OsvReport(TypedDict):
    results: List[Package]

def loadOsvReportJson(osvScannerStdout:str = None):
    osvReport: OsvReport = json.loads(osvScannerStdout)
    return osvReport
'''

def performOsvScannerAndReport(btConfiguration:dto.BTConfiguration=None,
                               osvScannerCmdParts:list=[]):
    #log.debug("performOsvScannerAndReport(): type(osvScannerCmdParts)=%s" % type(osvScannerCmdParts))
    # CSV file header
    csvFileHandle = open("%s/%s" % (btConfiguration.reportsDirectory, btConfiguration.osvCsvFileName), 'w')
    csvFileHandle.write('OSV URL,Ecosystem,Package,Version,Severity,Summary\n')
    # HTML file header
    htmlFileHandle = open("%s/%s" % (btConfiguration.reportsDirectory, btConfiguration.osvHtmlFileName), 'w')
    htmlFileHandle.write("<h1>OSV-Scanner Results</h1>\n")
    
    osvScannerRc, osvScannerStdout, osvScannerStderr = helpers.performDebugLogged(osvScannerCmdParts, collectOutput=True)
    if osvScannerRc == 0:
        log.debug("Got good RC %d from %s\nOSV-Scanner stderr:\n%s\n--- (end of osv-scanner stderr) ---\n" % (
            osvScannerRc, " ".join(osvScannerCmdParts), osvScannerStderr))
    else:
        log.warning("Got non-zero RC %d from OSV-Scanner, i.e. some vulnerability was found (or OSV-Scanner found no packages). Full command was: '%s\nOSV-Scanner stderr:\n%s\n--- (end of osv-scanner stderr) ---\n" % (
            osvScannerRc, " ".join(osvScannerCmdParts), osvScannerStderr))
        log.debug("OSV-Scanner stdout:\n%s\n--- (end of osv-scanner stdout) ---\n" % osvScannerStdout)
        #raise Exception("Got bad RC %d from OSV-Scanner, see above for details" % osvScannerRc)
        # rc 128 might mean "No package sources found" 
    
    if osvScannerStdout is None:
        log.error("OSV-Scanner report is None,see above for details")
        raise Exception("OSV-Scanner report is None, see above for details")
    if len(osvScannerStdout) == 0:
        log.error("OSV-Scanner report is empty,see above for details")
        raise Exception("OSV-Scanner report is empty, see above for details")

    #  parsing that vulnerability report and writing it to osv-report.csv, osv-report.html
    osvJsonDict = json.loads(osvScannerStdout)
    # { "results": [] }
    osvResults = osvJsonDict["results"]
    if osvResults is None:
        log.error("OSV-Scanner report contains no 'results' field:\n%s\n--- (end of osv-scanner output ---\n" % osvScannerStdout)
        raise Exception("OSV-Scanner report contains no 'results', see above for details")
    
    nrOsvVulnerablePackages = sum(len(r["packages"]) for r in osvResults)
    nrOsvVulnerabilities = 0
    if nrOsvVulnerablePackages == 0:
        log.notice("OSV-Scanner did not report any vulnerabilities")
        csvFileHandle.write('-,%s,-,-,No vulnerabilities found\n' % btConfiguration.defaultEcoSystem)
        htmlFileHandle.write("<p>No vulnerabilities found</p>\n")
    else:
        log.notice("OSV-Scanner found %d vulnerabilities, writing them to %s/%s and %s/%s" % (
            nrOsvVulnerablePackages, btConfiguration.reportsDirectory, btConfiguration.osvCsvFileName, btConfiguration.reportsDirectory, btConfiguration.osvHtmlFileName))
        htmlFileHandle.write("<p>Found vulnerabilities in %d packages:</p>\n" % nrOsvVulnerablePackages)
        if len(btConfiguration.htmlHints) > 1:
            htmlFileHandle.write(btConfiguration.htmlHints)
        htmlFileHandle.write('<table border="1" rules="all">\n  <thead><tr><th>OSV ID</th><th>Ecosystem</th><th>Package</th><th>Version</th><th>Severity</th><th>Summary</th></tr></thead>\n<tbody>\n')
        for osvResult in osvResults:
            for packageDict in osvResult["packages"]:
                packageMeta = packageDict["package"]
                packageName = packageMeta["name"]
                packageVersion = packageMeta["version"]
                packageEcoSystem = packageMeta["ecosystem"]
                for vulnerability in packageDict["vulnerabilities"]:
                    nrOsvVulnerabilities += 1
                    vulnerabilitySchema = vulnerability["schema_version"]
                    vulnerabilityId = vulnerability["id"]
                    if "summary" in vulnerability.keys():
                        vulnerabilitySummary = vulnerability["summary"]
                    elif "details" in vulnerability.keys():
                        vulnerabilitySummary = vulnerability["details"]
                    else:
                        vulnerabilitySummary = "- (no summary or details available)"
                    if vulnerabilitySchema[0:4] == "1.4.":
                        vulnerabilityUrl = "https://osv.dev/%s" % vulnerabilityId
                    else:
                        # in doubt attempt the newest one
                        vulnerabilityUrl = "https://osv.dev/%s" % vulnerabilityId
                    if "severity" in vulnerability.keys():
                        #Typically fou
                        sev = vulnerability["severity"]
                        if isinstance(sev, list):
                            #Sometimes values are in singleton lists
                            sev = sev[0]
                        vulnerabilitySeverity = helpers.getCvssScore(sev) or "Not specified"
                    # OSV URL,Ecosystem,Package,Version,Severity,Summary
                    csvFileHandle.write('%s,%s,%s,%s,%s,"%s"\n' % (
                        vulnerabilityUrl, packageEcoSystem, packageName, packageVersion, vulnerabilitySeverity, vulnerabilitySummary))
                    htmlFileHandle.write('  <tr><td><a href="%s">%s</a></td><td>%s</td><td>%s</td><td>%s</td><td>%s</td><td>%s</td></tr>\n' % (
                        vulnerabilityUrl, vulnerabilityId, packageEcoSystem, packageName, packageVersion, vulnerabilitySeverity, vulnerabilitySummary))

    csvFileHandle.close()
    # HTML file footer?
    if nrOsvVulnerablePackages > 0:
        htmlFileHandle.write("</tbody></table>\n")
        htmlFileHandle.write("<p>Listed %d vulnerabilities in %d packages.</p>\n" % (nrOsvVulnerabilities, nrOsvVulnerablePackages))
    if len(btConfiguration.htmlHints) > 1:
        htmlFileHandle.write(btConfiguration.htmlHints)
    htmlFileHandle.write("<p>Created by bomtools at %s</p>" % datetime.now().astimezone().isoformat(sep=" ", timespec="seconds"))
    htmlFileHandle.close()
            

